import asyncio
import logging
import math
from typing import Tuple, Any, Dict, List
import numpy as np
from datetime import datetime, timezone

from pyobs.interfaces import ICamera, IWindow, IBinning, ITemperatures, IAbortable
from pyobs.modules.camera.basecamera import BaseCamera
from pyobs.images import Image
from pyobs.utils.enums import ExposureStatus

# import the Andor SDK3 python wrapper provided with the camera ...
from .atcore import ATCore

log = logging.getLogger(__name__)


def _binning_string_to_tuple(bin):
    """
    Convert a binning string of the format 'axb' into a tuple (a, b) where a and b are integers.

    Parameters:
    bin (str): A string representing dimensions in the format 'axb', where a and b are integers.

    Returns:
    tuple: A tuple (a, b) where a and b are integers extracted from the input string.
    """
    # Split the string by 'x'
    a, b = bin.split("x")
    # Convert the split strings to integers and return as a tuple
    return int(a), int(b)


class AndorZylaCamera(
    BaseCamera, ICamera, IWindow, IBinning, ITemperatures, IAbortable
):
    """A pyobs module for Andor Zyla USB cameras."""

    __module__ = "pyobs_andors"

    def __init__(self, setpoint: float = 0.00, **kwargs: Any):
        """Initializes a new  AndorZylaCamera."""
        BaseCamera.__init__(self, **kwargs)

        # set the andor skd3 driver
        self._driver = ATCore()
        self._camera = None

        # activate lock
        self._lock_active = asyncio.Lock()

        # window and binning values init ...
        self._full_frame = (0, 0, 0, 0)
        self._window = (0, 0, 0, 0)
        self._binning = (1, 1)
        self._binnings_available = []

    async def open(self) -> None:
        """Open module.

        Raises:
            ValueError: If cannot connect to camera.
        """
        await BaseCamera.open(self)

        # open driver add pyobs utils exceptions ...
        log.info("Connecting to Andor camera ...")
        try:
            # connect to the Andor camera with device index 0 ...
            self._camera = self._driver.open(0)

            # Let user know we have a successful connection
            serial = self._driver.get_string(self._camera, "SerialNumber")
            model = self._driver.get_string(self._camera, "CameraModel")
            log.info("Connected to camera %s with serial number: %s", model, serial)

        except ValueError as e:
            raise ValueError("Could not establish link: %s" % str(e))

        # ask the camera for its available binning options and convert to a list of tuples ...
        self._binnings_available = [
            (int(binning.split("x")[0]), int(binning.split("x")[1]))
            for binning in self._driver.get_enum_string_options(
                self._camera, "AOIBinning"
            )
        ]

        # get the cameras initial full frame size ...
        self._full_frame = (
            0,
            0,
            self._driver.get_int(self._camera, "SensorWidth"),
            self._driver.get_int(self._camera, "SensorHeight"),
        )

        # get the cameras current area of interest window frame, noting that the SDK3 returns the coords shifted by +1 ...
        self._window = (
            self._driver.get_int(self._camera, "AOILeft") - 1,
            self._driver.get_int(self._camera, "AOITop") - 1,
            self._driver.get_int(self._camera, "AOIWidth"),
            self._driver.get_int(self._camera, "AOIHeight"),
        )

        # get the current binning settings of the camera ...
        self._binning = _binning_string_to_tuple(
            self._driver.get_enum_string(self._camera, "AOIBinning")
        )

        # Set to 16-bit (low noise & high well capacity)
        self._driver.set_enum_string(
            self._camera,
            "SimplePreAmpGainControl",
            "16-bit (low noise & high well capacity)",
        )

        # Set shutter mode to rolling
        self._driver.set_enum_string(
            self._camera, "ElectronicShutteringMode", "Rolling"
        )

        # Set readout to 200 MHz
        self._driver.set_enum_string(self._camera, "PixelReadoutRate", "280 MHz")

        # Turn off cooling on initialization
        self._driver.set_bool(self._camera, "SensorCooling", False)

        # Zyla cameras only a support single set point of 0°C. Cooler control is restricted to enabling or disabling the cooling ...
        # start cooling on camera init as not enough functionality to include ICooling interface ...
        # log.info("Enabling cooling and setting setpoint 0°C.")
        # self._driver.set_bool(self._camera, "SensorCooling", 1)

    async def close(self) -> None:
        """Close the module."""
        await BaseCamera.close(self)

        # turn off cooling and disconnect camera ...
        self._driver.set_bool(self._camera, "SensorCooling", 0)
        self._driver.close(self._camera)

    async def get_full_frame(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns full size of CMOS.

        Returns:
            Tuple with left, top, width, and height set.
        """
        return self._full_frame

    async def get_window(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns the camera window.

        Returns:
            Tuple with left, top, width, and height set.
        """
        return self._window

    async def set_window(
        self, left: int, top: int, width: int, height: int, **kwargs: Any
    ) -> None:
        """Set the camera window.

        Args:
            left: X offset of window.
            top: Y offset of window.
            width: Width of window.
            height: Height of window.

        Raises:
            ValueError: If binning could not be set.
        """
        self._window = (left, top, width, height)
        log.info("Setting window to %dx%d at %d,%d...", width, height, left, top)

    async def get_binning(self, **kwargs: Any) -> Tuple[int, int]:
        """Returns the camera binning.

        Returns:
            Tuple with x and y.
        """
        return self._binning

    async def set_binning(self, x: int, y: int, **kwargs: Any) -> None:
        """Set the camera binning.

        Args:
            x: X binning.
            y: Y binning.

        Raises:
            ValueError: If binning could not be set.
        """
        self._binning = (x, y)
        log.info("Setting binning to %dx%d...", x, y)

    async def list_binnings(self, **kwargs: Any) -> List[Tuple[int, int]]:
        """List available binnings.

        Returns:
            List of available binnings as (x, y) tuples.
        """
        return self._binnings_available

    async def get_temperatures(self, **kwargs: Any) -> Dict[str, float]:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """
        return {
            "CMOS": self._driver.get_float(self._camera, "SensorTemperature"),
        }

    async def _expose(
        self, exposure_time: float, open_shutter: bool, abort_event: asyncio.Event
    ) -> Image:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            exposure_time: The requested exposure time in seconds.
            open_shutter: Whether or not to open the shutter.
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image.

        Raises:
            GrabImageError: If exposure was not successful.
        """

        # set the binning on the camera ...
        log.info("Set binning to %dx%d.", self._binning[0], self._binning[1])
        self._driver.set_enum_string(
            self._camera, "AOIBinning", f"{self._binning[0]}x{self._binning[1]}"
        )

        # determine window, noting from the SDK3 documentation: "The AOIWidth and AOIHeight features are set and retrieved in
        # units of super-pixels Therefore, when binning is in use, the AOIWidth value will always indicate the number of data
        # pixels that each row of the image data contains and not the number of pixels read off the sensor. The AOILeft and
        # AOITop coordinates are specified in units of sensor pixels."
        width = int(math.floor(self._window[2]) / self._binning[0])
        height = int(math.floor(self._window[3]) / self._binning[1])

        log.info(
            "Set window to %dx%d (binned %dx%d) at %d,%d.",
            self._window[2],
            self._window[3],
            width,
            height,
            self._window[0],
            self._window[1],
        )

        # set the camera encoding ...
        self._driver.set_enum_string(self._camera, "PixelEncoding", "Mono16")

        # set the window on the camera remembering the SDK3 has the origin at (+1,+1) ...
        self._driver.set_int(self._camera, "AOIWidth", width)
        self._driver.set_int(self._camera, "AOIHeight", height)
        self._driver.set_int(self._camera, "AOILeft", self._window[0] + 1)
        self._driver.set_int(self._camera, "AOITop", self._window[1] + 1)

        # set the cameras shutter state ...
        if open_shutter:
            self._driver.set_enum_string(self._camera, "ShutterMode", "Open")
        else:
            self._driver.set_enum_string(self._camera, "ShutterMode", "Closed")

        # set the cameras exposure time in seconds ...
        self._driver.set_float(self._camera, "ExposureTime", exposure_time)

        # get the expected size of the required memory buffer in bytes from the camera ...
        imageSizeBytes = self._driver.get_int(self._camera, "ImageSizeBytes")

        # setup a primary and secondary buffer as per SDK3 documentation ...s
        primary_buffer = np.empty((imageSizeBytes,), dtype="B")
        self._driver.queue_buffer(
            self._camera, primary_buffer.ctypes.data, imageSizeBytes
        )

        secondary_buffer = np.empty((imageSizeBytes,), dtype="B")
        self._driver.queue_buffer(
            self._camera, secondary_buffer.ctypes.data, imageSizeBytes
        )

        # get the current date and time of exposure ...
        date_obs = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%f")

        log.info(
            "Starting exposure with %s shutter for %.2f seconds...",
            "open" if open_shutter else "closed",
            exposure_time,
        )

        # start the exposure ...
        self._driver.command(self._camera, "AcquisitionStart")

        # wait for exposure to finish or be aborted before reading out ...
        await self._wait_exposure(abort_event, exposure_time, open_shutter)

        # begin the readout ...
        log.info("Exposure finished, reading out...")
        await self._change_exposure_status(ExposureStatus.READOUT)
        width = int(math.floor(self._window[2] / self._binning[0]))
        height = int(math.floor(self._window[3] / self._binning[1]))

        # we need the stride data ... from SDK3 documentation: "AOIStride - At the end of each row there may be additional padding bytes.
        # This padding area does not contain any valid pixel data and should be skipped over when processing or displaying an image."
        aoistride = self._driver.get_int(self._camera, "AOIStride")

        # read in the data from the primary buffer and remove stride/padding pixels  ...
        primary_buffer = primary_buffer[0 : height * aoistride]
        np_d = primary_buffer.view(dtype=np.uint16)
        np_d = np_d.reshape(height, round(np_d.size / height))
        img = np_d[0:height, 0:width]

        # create FITS image and set header ...
        image = Image(img)
        image.header["DATE-OBS"] = (date_obs, "Date and time of start of exposure")
        image.header["EXPTIME"] = (exposure_time, "Exposure time [s]")
        image.header["DET-TEMP"] = (
            self._driver.get_float(self._camera, "SensorTemperature"),
            "CMOS temperature [C]",
        )

        # camera has fixed setpoint ...
        image.header["DET-TSET"] = (0, "Cooler setpoint [C]")

        # note instrument and detector ...
        image.header["INSTRUMENT"] = (
            self._driver.get_string(self._camera, "CameraModel"),
            "Name of instrument",
        )

        # note binning ...
        image.header["XBINNING"] = image.header["DET-BIN1"] = (
            self._binning[0],
            "Binning factor used on X axis",
        )
        image.header["YBINNING"] = image.header["DET-BIN2"] = (
            self._binning[1],
            "Binning factor used on Y axis",
        )

        # note window ...
        image.header["XORGSUBF"] = (self._window[0], "Subframe origin on X axis")
        image.header["YORGSUBF"] = (self._window[1], "Subframe origin on Y axis")

        # note statistics ...
        image.header["DATAMIN"] = (float(np.min(img)), "Minimum data value")
        image.header["DATAMAX"] = (float(np.max(img)), "Maximum data value")
        image.header["DATAMEAN"] = (float(np.mean(img)), "Mean data value")

        # biassec/trimsec
        self.set_biassec_trimsec(image.header, *self._full_frame)

        # return FITS image
        log.info("Readout finished.")
        return image

    async def _wait_exposure(
        self, abort_event: asyncio.Event, exposure_time: float, open_shutter: bool
    ) -> None:
        """Wait for exposure to finish.

        Params:
            abort_event: Event that aborts the exposure.
            exposure_time: Exp time in sec.
        """

        while True:
            # has the exposure been aborted?
            if abort_event.is_set():
                await self._change_exposure_status(ExposureStatus.IDLE)
                raise InterruptedError("Aborted exposure.")

            # is exposure finished?
            try:
                self._driver.wait_buffer(self._camera, timeout=0)
                self._driver.command(self._camera, "AcquisitionStop")
                break
            except:
                await asyncio.sleep(0.01)

    async def _abort_exposure(self) -> None:
        """Abort the running exposure. Should be implemented by derived class.

        Raises:
            ValueError: If an error occured.
        """
        self._driver.command(self._camera, "AcquisitionStop")


__all__ = ["AndorZylaCamera"]



cimport numpy as np
import numpy as np


cdef extern from "atmcdLXd.h":
    unsigned int CoolerOFF();
    unsigned int CoolerON();
    unsigned int GetAcquiredData16(unsigned short *arr, int size)
    unsigned int GetDetector(int *xpixels, int *ypixels)
    unsigned int GetStatus(int *status)
    unsigned int GetTemperatureF(float *temperature)
    unsigned int Initialize(char *dir)
    unsigned int SetAcquisitionMode(int mode)
    unsigned int SetExposureTime(float time)
    unsigned int SetImage(int hbin, int vbin, int hstart, int hend, int vstart, int vend)
    unsigned int SetReadMode(int mode)
    unsigned int SetShutter(int typ, int mode, int closingtime, int openingtime)
    unsigned int SetTemperature(int temperature)
    unsigned int ShutDown()
    unsigned int StartAcquisition();
    unsigned int WaitForAcquisition();
  
    int DRV_SUCCESS
    int DRV_NOT_INITIALIZED
    int DRV_ACQUIRING 
    int DRV_ERROR_ACK
    int DRV_NOT_SUPPORTED
    int DRV_P1INVALID
    int DRV_P2INVALID
    int DRV_P3INVALID
    int DRV_P4INVALID
    int DRV_P5INVALID
    int DRV_P6INVALID
    int DRV_VXDNOTINSTALLED
    int DRV_INIERROR
    int DRV_COFERROR
    int DRV_FLEXERROR
    int DRV_ERROR_FILELOAD
    int DRV_ERROR_PAGELOCK
    int DRV_USBERROR
    int DRV_ERROR_NOCAMERA
    int DRV_NO_NEW_DATA
    int DRV_ACQUISITION_ERRORS
    int DRV_INVALID_FILTER
    int DRV_BINNING_ERROR
    int DRV_SPOOLSETUPERROR
    int DRV_IDLE
    int DRV_TEMPCYCLE
    int DRV_ACCUM_TIME_NOT_MET
    int DRV_KINETIC_TIME_NOT_MET
    int DRV_ACQ_BUFFER
    int DRV_ACQ_DOWNFIFO_FULL
    int DRV_SPOOLERROR
    int DRV_TEMP_OFF
    int DRV_TEMP_STABILIZED
    int DRV_TEMP_NOT_REACHED
    int DRV_TEMP_DRIFT
    int DRV_TEMP_NOT_STABILIZED
 

ANDOR_ACQUIRING = DRV_ACQUIRING
ANDOR_TEMP_OFF = DRV_TEMP_OFF
ANDOR_TEMP_STABILIZED = DRV_TEMP_STABILIZED
ANDOR_TEMP_NOT_REACHED = DRV_TEMP_NOT_REACHED
ANDOR_TEMP_DRIFT = DRV_TEMP_DRIFT
ANDOR_TEMP_NOT_STABILIZED = DRV_TEMP_NOT_STABILIZED

class AndorException(Exception):
    pass


def coolerOn():
    # call library
    res = CoolerON()
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
            }[res])


def coolerOff():
    # call library
    res = CoolerOFF()
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
            }[res])
    

def getAcquiredData(width, height):
    # create numpy array of given dimensions
    cdef np.ndarray[unsigned short, ndim=1] image = np.zeros((width * height), dtype=np.ushort)
    
    # call library
    res = GetAcquiredData16(<unsigned short*>image.data, width*height);
    print(res)
    
        # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
                DRV_P1INVALID: 'Invalid pointer (i.e. NULL).',
                DRV_P2INVALID: 'Array size isincorrect.',
                DRV_NO_NEW_DATA: 'No acquisition has taken place'
            }[res])
    
    # return image
    return image.reshape((height, width))

def getDetector():
    # define x and y
    cdef int x = 0
    cdef int y = 0
    
    # call library
    res = GetDetector(&x, &y)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.'
            }[res])
    
    # no error, return detector size
    return x, y


def getStatus():
    # define status
    cdef int status = 0
    
    # call library
    res = GetStatus(&status)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.'
            }[res])
    
    # no error, return status
    return status


def getTemperature():
    # define temp
    cdef float temp = 0
    
    # call library
    res = GetTemperatureF(&temp)
    
    # error checking
    if res in [DRV_NOT_INITIALIZED, DRV_ACQUIRING, DRV_ERROR_ACK]:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_ERROR_ACK: 'Unable to communicate with card.'
            }[res])
    
    # no error, return temperature and status
    return temp, res

def init(dir):
    # call library
    res = Initialize(bytes(dir, 'utf8'))
     
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_VXDNOTINSTALLED: 'VxD not loaded.',
                DRV_INIERROR: 'Unable to load "DETECTOR.INI".',
                DRV_COFERROR: 'Unable to load “*.COF".',
                DRV_FLEXERROR: 'Unable to load “*.RBF".',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
                DRV_ERROR_FILELOAD: 'Unable to load “*.COF" or “*.RBF" files.',
                DRV_ERROR_PAGELOCK: 'Unable to acquire lock on requested memory.',
                DRV_USBERROR: 'Unable to detect USB device or not USB2.0.',
                DRV_ERROR_NOCAMERA: 'No camera found'
            }[res])


def setAcquisitionMode(mode):
    # call library
    res = SetAcquisitionMode(mode)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_P1INVALID: ' Acquisition Mode invalid.'
            }[res])
    
    
def setExposureTime(time):
    # call library
    res = SetExposureTime(time)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_P1INVALID: 'Exposure Time invalid.'
            }[res])
    
    
def setImage(hbin, vbin, hstart, hend, vstart, vend):
    # call library
    res = SetImage(hbin, vbin, hstart, hend, vstart, vend)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_P1INVALID: 'Binning parameters invalid.',
                DRV_P2INVALID: 'Binning parameters invalid.',
                DRV_P3INVALID: 'Sub-area co-ordinate is invalid.',
                DRV_P4INVALID: 'Sub-area co-ordinate is invalid.',
                DRV_P5INVALID: 'Sub-area co-ordinate is invalid.',
                DRV_P6INVALID: 'Sub-area co-ordinate is invalid.'
            }[res])

    
def setReadMode(mode):
    # call library
    res = SetReadMode(mode)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_P1INVALID: 'Invalid readout mode passed.'
            }[res])

def setShutter(typ, mode, closingtime, openingtime):
    # call library
    res = SetShutter(typ, mode, closingtime, openingtime)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
                DRV_NOT_SUPPORTED: 'Camera does not support shutter control.',
                DRV_P1INVALID: 'Invalid TTL type.',
                DRV_P2INVALID: 'Invalid mode.',
                DRV_P3INVALID: 'Invalid time to open.',
                DRV_P4INVALID: 'Invalid time to close.'
            }[res])


def setTemperature(temp):
    # call library
    res = SetTemperature(temp)
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
                DRV_P1INVALID: 'Temperature invalid.',
                DRV_NOT_SUPPORTED: ' The camera does not support setting the temperature.'
            }[res])
    

def startAcquisition():
    # call library
    res = StartAcquisition()
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_ACQUIRING: 'Acquisition in progress.',
                DRV_VXDNOTINSTALLED: 'VxD not loaded.',
                DRV_ERROR_ACK: 'Unable to communicate with card.',
                DRV_INIERROR: 'Error reading “DETECTOR.INI".',
                DRV_ACQUISITION_ERRORS: 'Acquisition settings invalid.',
                DRV_ERROR_PAGELOCK: 'Unable to allocate memory.',
                DRV_INVALID_FILTER: 'Filter not available for current acquisition.',
                DRV_BINNING_ERROR: 'Range not multiple of horizontal binning.',
                DRV_SPOOLSETUPERROR: 'Error with spool settings.'
            }[res])


def shutdown():
    ShutDown()


def waitForAcquisition():
    # call library
    res = WaitForAcquisition()
    
    # error checking
    if res != DRV_SUCCESS:
        raise AndorException({
                DRV_NOT_INITIALIZED: 'System not initialized.',
                DRV_NO_NEW_DATA: 'Non-Acquisition Event occured.',
            }[res])



import logging
import math
import threading
from datetime import datetime
import time
import numpy as np
from astropy.io import fits

from pytel.interfaces import ICamera, ICameraWindow, ICameraBinning, ICooling
from pytel.modules.camera.basecamera import BaseCamera, CameraException
from .libandor import Initialize, Shutdown


log = logging.getLogger(__name__)


class AndorCamera(BaseCamera, ICamera, ICameraWindow, ICameraBinning, ICooling):
    def __init__(self, *args, **kwargs):
        BaseCamera.__init__(self, *args, **kwargs)

        # variables
        self._dev_domain = None
        self._dev_file = None
        self._dev_name = None
        self._connected = False
        self._dev = None
        self._temp_setpoint = None

        # window and binning
        self._window = None
        self._binning = None
        self._window_dirty = True

    def open(self) -> bool:
        # open base
        if not BaseCamera.open(self):
            return False

        # already open?
        if self._dev is not None:
            return False

        # create list of USB cameras
        fli.CreateList(fli.FLIDOMAIN_USB | fli.FLIDEVICE_CAMERA)

        # fetch first
        # return values are something like: (258 /dev/fliusb1 ProLine PL230)
        self._dev_domain, self._dev_file, self._dev_name = fli.ListFirst()

        # clean up list
        fli.DeleteList()

        # open connection
        log.info('Opening connection to "%s" at %s...', self._dev_name, self._dev_file)
        self._dev = fli.Open(self._dev_file, self._dev_domain)

        # get window and binning from camera
        self._window = self._get_window()
        self._binning = self._get_binning()
        self._window_dirty = False

        # set cooling
        self._temp_setpoint = None
        self.set_cooling(True, self.config['setpoint'])

        # success
        return True

    def close(self):
        # close base
        BaseCamera.close(self)

        # not open?
        if self._dev is not None:
            # close connection
            fli.Close(self._dev)
            self._dev = None

    def _check_dev(self):
        # not connected?
        if self._dev is None:
            raise CameraException("Not connected to camera.")

    @classmethod
    def default_config(cls):
        cfg = super(FliCamera, cls).default_config()
        cfg['setpoint'] = -20.
        return cfg

    def get_full_frame(self, *args, **kwargs) -> dict:
        self._check_dev()
        size = fli.GetVisibleArea(self._dev)
        return {'left': size[0], 'top': size[1], 'width': size[2]-size[0], 'height': size[3]-size[1]}

    def _get_window(self) -> dict:
        self._check_dev()
        dim = fli.GetReadoutDimensions(self._dev)
        return {'left': dim[1], 'top': dim[4], 'width': dim[0], 'height': dim[3]}

    def _get_binning(self) -> dict:
        self._check_dev()
        dim = fli.GetReadoutDimensions(self._dev)
        return {'x': dim[2], 'y': dim[5]}

    def get_window(self, *args, **kwargs) -> dict:
        return self._window

    def get_binning(self, *args, **kwargs) -> dict:
        return self._binning

    def set_window(self, left: int, top: int, width: int, height: int, *args, **kwargs) -> bool:
        self._window = {'left': int(left), 'top': int(top), 'width': int(width), 'height': int(height)}
        self._window_dirty = True
        log.info('Setting window to %dx%d at %d,%d...', width, height, left, top)
        return True

    def set_binning(self, x: int, y: int, *args, **kwargs) -> bool:
        self._binning = {'x': int(x), 'y': int(y)}
        self._window_dirty = True
        log.info('Setting binning to %dx%d...', x, y)
        return True

    def _set_window_binning(self):
        if self._window_dirty:
            # check device
            self._check_dev()

            # set binning
            log.info("Set binning to %dx%d.", self._binning['x'], self._binning['y'])
            fli.SetHBin(self._dev, int(self._binning['x']))
            fli.SetVBin(self._dev, int(self._binning['y']))

            # set window, divide width/height by binning, from libfli:
            # "Note that the given lower-right coordinate must take into account the horizontal and
            # vertical bin factor settings, but the upper-left coordinate is absolute."
            width = int(math.floor(self._window['width']) / self._binning['x'])
            height = int(math.floor(self._window['height']) / self._binning['y'])
            log.info("Set window to %dx%d (binned %dx%d) at %d,%d.",
                         self._window['width'], self._window['height'],
                         width, height,
                         self._window['left'], self._window['top'])
            fli.SetImageArea(self._dev, int(self._window['left']), int(self._window['top']),
                             int(self._window['left'] + width), int(self._window['top'] + height))

            # cleaned!
            self._window_dirty = False

    def _expose(self, exposure_time: int, open_shutter: bool, abort_event: threading.Event) -> fits.ImageHDU:
        self._check_dev()

        # set window/binning, if necessary
        self._set_window_binning()

        # set some stuff
        self._camera_status = ICamera.CameraStatus.EXPOSING
        fli.SetTDI(self._dev, 0, 0)
        fli.SetExposureTime(self._dev, int(exposure_time))
        fli.SetFrameType(self._dev, fli.FLI_FRAME_TYPE_NORMAL if open_shutter else fli.FLI_FRAME_TYPE_DARK)

        # get date obs
        log.info('Starting exposure with %s shutter for %.2f seconds...',
                 'open' if open_shutter else 'closed', exposure_time / 1000.)
        date_obs = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")

        # do exposure
        fli.ExposeFrame(self._dev)

        # wait for exposure to finish
        while True:
            # aborted?
            if abort_event.is_set():
                log.warning('Aborted exposure.')
                return None

            # get status
            camera_status = fli.GetDeviceStatus(self._dev)
            remaining_exposure = fli.GetExposureStatus(self._dev)

            # check it
            if (camera_status == fli.FLI_CAMERA_STATUS_UNKNOWN and remaining_exposure == 0) or \
                    (camera_status != fli.FLI_CAMERA_STATUS_UNKNOWN and camera_status & fli.FLI_CAMERA_DATA_READY):
                break
            else:
                # sleep a little
                time.sleep(0.2)

        # readout
        log.info('Exposure finished, reading out...')
        self._camera_status = ICamera.CameraStatus.READOUT
        width = int(math.floor(self._window['width'] / self._binning['x']))
        height = int(math.floor(self._window['height'] / self._binning['y']))
        img = np.zeros((height, width), dtype=np.uint16)
        for row in range(height):
            img[row, :] = fli.GrabRow(self._dev, width)

        # create FITS image and set header
        hdu = fits.PrimaryHDU(img)
        hdu.header['DATE-OBS'] = (date_obs, 'Date and time of start of exposure')
        hdu.header['EXPTIME'] = (exposure_time / 1000., 'Exposure time [s]')
        hdu.header['DET-TEMP'] = (fli.ReadTemperature(self._dev, fli.FLI_TEMPERATURE_CCD), 'CCD temperature [C]')
        hdu.header['DET-COOL'] = (fli.GetCoolerPower(self._dev), 'Cooler power [percent]')
        hdu.header['DET-TSET'] = (self._temp_setpoint, 'Cooler setpoint [C]')

        # instrument and detector
        hdu.header['INSTRUME'] = (self._dev_name, 'Name of instrument')

        # binning
        hdu.header['XBINNING'] = hdu.header['DET-BIN1'] = (self._binning['x'], 'Binning factor used on X axis')
        hdu.header['YBINNING'] = hdu.header['DET-BIN2'] = (self._binning['y'], 'Binning factor used on Y axis')

        # window
        hdu.header['XORGSUBF'] = (self._window['left'], 'Subframe origin on X axis')
        hdu.header['YORGSUBF'] = (self._window['top'], 'Subframe origin on Y axis')

        # statistics
        hdu.header['DATAMIN'] = (float(np.min(img)), 'Minimum data value')
        hdu.header['DATAMAX'] = (float(np.max(img)), 'Maximum data value')
        hdu.header['DATAMEAN'] = (float(np.mean(img)), 'Mean data value')

        # biassec/trimsec
        full = self.get_full_frame()
        self.set_biassec_trimsec(hdu.header, full['left'], full['top'], full['width'], full['height'])

        # return FITS image
        log.info('Readout finished.')
        self._camera_status = ICamera.CameraStatus.IDLE
        return hdu

    def _abort_exposure(self) -> bool:
        """Aborts the current exposure.

        Returns:
            bool: True if successful, otherwise False.
        """
        self._check_dev()
        fli.CancelExposure(self._dev)
        self._camera_status = ICamera.CameraStatus.IDLE
        return True

    def status(self, *args, **kwargs) -> dict:
        self._check_dev()

        # get status from parent
        s = super().status()

        # add cooling stuff
        s['ICooling'] = {
            'Enabled': self._temp_setpoint is not None,
            'SetPoint': self._temp_setpoint,
            'Power': fli.GetCoolerPower(self._dev),
            'Temperatures': {
                'CCD': fli.ReadTemperature(self._dev, fli.FLI_TEMPERATURE_CCD),
                'Base': fli.ReadTemperature(self._dev, fli.FLI_TEMPERATURE_BASE)
            }
        }

        # finished
        return s

    def set_cooling(self, enabled: bool, setpoint: float, *args, **kwargs) -> bool:
        self._check_dev()

        # log
        if enabled:
            log.info('Enabling cooling with a setpoint of %.2f°C...', setpoint)
        else:
            log.info('Disabling cooling and setting setpoint to 20°C...')

        # if not enabled, set setpoint to None
        self._temp_setpoint = setpoint if enabled else None

        # set setpoint
        fli.SetTemperature(self._dev, float(setpoint) if setpoint is not None else 20.)
        return True



import sys
if sys.version < '3':
    import codecs
    def u(x):
        return codecs.unicode_escape_decode(x)[0]
else:
    def u(x):
        return x

class ATCoreException(Exception): pass

class ATCore(object):

    _ERRORS = {
        0: 'AT_SUCCESS',
        1: 'AT_ERR_NOTINITIALISED',
        2: 'AT_ERR_NOTIMPLEMENTED',
        3: 'AT_ERR_READONLY',
        4: 'AT_ERR_NOTREADABLE',
        5: 'AT_ERR_NOTWRITABLE',
        6: 'AT_ERR_OUTOFRANGE',
        7: 'AT_ERR_INDEXNOTAVAILABLE',
        8: 'AT_ERR_INDEXNOTIMPLEMENTED',
        9: 'AT_ERR_EXCEEDEDMAXSTRINGLENGTH',
        10: 'AT_ERR_CONNECTION',
        11: 'AT_ERR_NODATA',
        12: 'AT_ERR_INVALIDHANDLE',
        13: 'AT_ERR_TIMEDOUT',
        14: 'AT_ERR_BUFFERFULL',
        15: 'AT_ERR_INVALIDSIZE',
        16: 'AT_ERR_INVALIDALIGNMENT',
        17: 'AT_ERR_COMM',
        18: 'AT_ERR_STRINGNOTAVAILABLE',
        19: 'AT_ERR_STRINGNOTIMPLEMENTED',
        20: 'AT_ERR_NULL_FEATURE',
        21: 'AT_ERR_NULL_HANDLE',
        22: 'AT_ERR_NULL_IMPLEMENTED_VAR',
        23: 'AT_ERR_NULL_READABLE_VAR',
        24: 'AT_ERR_NULL_READONLY_VAR',
        25: 'AT_ERR_NULL_WRITABLE_VAR',
        26: 'AT_ERR_NULL_MINVALUE',
        27: 'AT_ERR_NULL_MAXVALUE',
        28: 'AT_ERR_NULL_VALUE',
        29: 'AT_ERR_NULL_STRING',
        30: 'AT_ERR_NULL_COUNT_VAR',
        31: 'AT_ERR_NULL_ISAVAILABLE_VAR',
        32: 'AT_ERR_NULL_MAXSTRINGLENGTH',
        33: 'AT_ERR_NULL_EVCALLBACK',
        34: 'AT_ERR_NULL_QUEUE_PTR',
        35: 'AT_ERR_NULL_WAIT_PTR',
        36: 'AT_ERR_NULL_PTRSIZE',
        37: 'AT_ERR_NOMEMORY',
        38: 'AT_AT_ERR_DEVICEINUSE',
        100: 'AT_ERR_HARDWARE_OVERFLOW',
    }
    __version__ = '0.1'
    LIBRARY_NAME = 'atcore'
    AT_INFINITE=0xFFFFFFFF

    AT_CALLBACK_SUCCESS = 0

    AT_TRUE = 1
    AT_FALSE = 0

    AT_SUCCESS = 0
    AT_ERR_NOTINITIALISED = 1
    AT_ERR_NOTIMPLEMENTED = 2
    AT_ERR_READONLY = 3
    AT_ERR_NOTREADABLE = 4
    AT_ERR_NOTWRITABLE = 5
    AT_ERR_OUTOFRANGE= 6
    AT_ERR_INDEXNOTAVAILABLE= 7
    AT_ERR_INDEXNOTIMPLEMENTED= 8
    AT_ERR_EXCEEDEDMAXSTRINGLENGTH= 9
    AT_ERR_CONNECTION= 10
    AT_ERR_NODATA= 11
    AT_ERR_INVALIDHANDLE= 12
    AT_ERR_TIMEDOUT= 13
    AT_ERR_BUFFERFULL= 14
    AT_ERR_INVALIDSIZE= 15
    AT_ERR_INVALIDALIGNMENT= 16
    AT_ERR_COMM= 17
    AT_ERR_STRINGNOTAVAILABLE= 18
    AT_ERR_STRINGNOTIMPLEMENTED= 19

    AT_ERR_NULL_FEATURE= 20
    AT_ERR_NULL_HANDLE= 21
    AT_ERR_NULL_IMPLEMENTED_VAR= 22
    AT_ERR_NULL_READABLE_VAR= 23
    AT_ERR_NULL_READONLY_VAR= 24
    AT_ERR_NULL_WRITABLE_VAR= 25
    AT_ERR_NULL_MINVALUE= 26
    AT_ERR_NULL_MAXVALUE= 27
    AT_ERR_NULL_VALUE= 28
    AT_ERR_NULL_STRING= 29
    AT_ERR_NULL_COUNT_VAR= 30
    AT_ERR_NULL_ISAVAILABLE_VAR= 31
    AT_ERR_NULL_MAXSTRINGLENGTH= 32
    AT_ERR_NULL_EVCALLBACK= 33
    AT_ERR_NULL_QUEUE_PTR= 34
    AT_ERR_NULL_WAIT_PTR= 35
    AT_ERR_NULL_PTRSIZE= 36
    AT_ERR_NOMEMORY= 37
    AT_ERR_DEVICEINUSE= 38
    AT_ERR_DEVICENOTFOUND= 39

    AT_ERR_HARDWARE_OVERFLOW= 100

    AT_HANDLE_UNINITIALISED= -1
    AT_HNDL_SYSTEM = 1

    def __init__(self):
        from cffi import FFI
        self.ffi = FFI()
        self.ffi.set_unicode(True)
        self.C = self.ffi.cdef("""
        typedef int AT_H;
        typedef int AT_BOOL;
        typedef long long AT_64;
        typedef unsigned char AT_U8;
        typedef wchar_t AT_WC;

        int AT_InitialiseLibrary();
        int AT_FinaliseLibrary();

        int AT_Open(int CameraIndex, AT_H *Hndl);
        int AT_Close(AT_H Hndl);

        int AT_IsImplemented(AT_H Hndl, const AT_WC* Feature, AT_BOOL* Implemented);
        int AT_IsReadable(AT_H Hndl, const AT_WC* Feature, AT_BOOL* Readable);
        int AT_IsWritable(AT_H Hndl, const AT_WC* Feature, AT_BOOL* Writable);
        int AT_IsReadOnly(AT_H Hndl, const AT_WC* Feature, AT_BOOL* ReadOnly);

        int AT_SetInt(AT_H Hndl, const AT_WC* Feature, AT_64 Value);
        int AT_GetInt(AT_H Hndl, const AT_WC* Feature, AT_64* Value);
        int AT_GetIntMax(AT_H Hndl, const AT_WC* Feature, AT_64* MaxValue);
        int AT_GetIntMin(AT_H Hndl, const AT_WC* Feature, AT_64* MinValue);

        int AT_SetFloat(AT_H Hndl, const AT_WC* Feature, double Value);
        int AT_GetFloat(AT_H Hndl, const AT_WC* Feature, double* Value);
        int AT_GetFloatMax(AT_H Hndl, const AT_WC* Feature, double* MaxValue);
        int AT_GetFloatMin(AT_H Hndl, const AT_WC* Feature, double* MinValue);

        int AT_SetBool(AT_H Hndl, const AT_WC* Feature, AT_BOOL Value);
        int AT_GetBool(AT_H Hndl, const AT_WC* Feature, AT_BOOL* Value);

        /*
        int AT_SetEnumerated(AT_H Hndl, const AT_WC* Feature, int Value);
        int AT_SetEnumeratedString(AT_H Hndl, const AT_WC* Feature, const AT_WC* String);
        int AT_GetEnumerated(AT_H Hndl, const AT_WC* Feature, int* Value);
        int AT_GetEnumeratedCount(AT_H Hndl,const  AT_WC* Feature, int* Count);
        int AT_IsEnumeratedIndexAvailable(AT_H Hndl, const AT_WC* Feature, int Index, AT_BOOL* Available);
        int AT_IsEnumeratedIndexImplemented(AT_H Hndl, const AT_WC* Feature, int Index, AT_BOOL* Implemented);
        int AT_GetEnumeratedString(AT_H Hndl, const AT_WC* Feature, int Index, AT_WC* String, int StringLength);
        */

        int AT_SetEnumIndex(AT_H Hndl, const AT_WC* Feature, int Value);
        int AT_SetEnumString(AT_H Hndl, const AT_WC* Feature, const AT_WC* String);
        int AT_GetEnumIndex(AT_H Hndl, const AT_WC* Feature, int* Value);
        int AT_GetEnumCount(AT_H Hndl,const  AT_WC* Feature, int* Count);
        int AT_IsEnumIndexAvailable(AT_H Hndl, const AT_WC* Feature, int Index, AT_BOOL* Available);
        int AT_IsEnumIndexImplemented(AT_H Hndl, const AT_WC* Feature, int Index, AT_BOOL* Implemented);
        int AT_GetEnumStringByIndex(AT_H Hndl, const AT_WC* Feature, int Index, AT_WC* String, int StringLength);

        int AT_Command(AT_H Hndl, const AT_WC* Feature);

        int AT_SetString(AT_H Hndl, const AT_WC* Feature, const AT_WC* String);
        int AT_GetString(AT_H Hndl, const AT_WC* Feature, AT_WC* String, int StringLength);
        int AT_GetStringMaxLength(AT_H Hndl, const AT_WC* Feature, int* MaxStringLength);

        int AT_QueueBuffer(AT_H Hndl, AT_U8* Ptr, int PtrSize);
        int AT_WaitBuffer(AT_H Hndl, AT_U8** Ptr, int* PtrSize, unsigned int Timeout);
        int AT_Flush(AT_H Hndl);

        """)
        self.lib = self.ffi.verify('#include "atcore.h"', include_dirs=["."], libraries=["atcore"])
        self.handle_return(self.lib.AT_InitialiseLibrary())

    def __del__(self):
        self.handle_return(self.lib.AT_FinaliseLibrary())

    def handle_return(self,ret_value):
        if ret_value != 0:
            raise ATCoreException('{} ({})'.format(ret_value, self._ERRORS[ret_value]))
        return ret_value

    def get_version(self):
        return self.__version__

    def open(self, index):
        """Open camera AT_H.
        """
        result = self.ffi.new("AT_H *")
        self.handle_return(self.lib.AT_Open(index, result))

        return result[0]

    def close(self, AT_H):
        """Close camera AT_H.
        """
        self.handle_return(self.lib.AT_Close(AT_H))

    def is_implemented(self, AT_H, command):
        """Checks if command is implemented.
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_IsImplemented(AT_H, u(command), result))
        return result[0]

    def is_readable(self, AT_H, command):
        """Checks if command is readable.
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_IsReadable(AT_H, u(command), result))
        return result[0]

    def is_writable(self, AT_H, command):
        """Checks if command is writable.
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_IsWritable(AT_H, u(command), result))
        return result[0]

    def is_readonly(self, AT_H, command):
        """Checks if command is read only.
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_IsReadOnly(AT_H, u(command), result))
        return result[0]

    def set_int(self, AT_H, command, value):
        """SetInt function.
        """
        self.handle_return(self.lib.AT_SetInt(AT_H, u(command), value))

    def get_int(self, AT_H, command):
        """Run command and get Int return value.
        """
        result = self.ffi.new("AT_64 *")
        self.handle_return(self.lib.AT_GetInt(AT_H, u(command), result))
        return result[0]

    def get_int_max(self, AT_H, command):
        """Run command and get maximum Int return value.
        """
        result = self.ffi.new("AT_64 *")
        self.handle_return(self.lib.AT_GetIntMax(AT_H, u(command), result))
        return result[0]

    def get_int_min(self, AT_H, command):
        """Run command and get minimum Int return value.
        """
        result = self.ffi.new("AT_64 *")
        self.handle_return(self.lib.AT_GetIntMin(AT_H, u(command), result))
        return result[0]

    def set_float(self, AT_H, command, value):
        """Set command with Float value parameter.
        """
        self.handle_return(self.lib.AT_SetFloat(AT_H, u(command), value))

    def get_float(self, AT_H, command):
        """Run command and get float return value.
        """
        result = self.ffi.new("double *")
        self.handle_return(self.lib.AT_GetFloat(AT_H, u(command), result))
        return result[0]

    def get_float_max(self, AT_H, command):
        """Run command and get maximum float return value.
        """
        result = self.ffi.new("double *")
        self.handle_return(self.lib.AT_GetFloatMax(AT_H, u(command), result))
        return result[0]

    def get_float_min(self, AT_H, command):
        """Run command and get minimum float return value.
        """
        result = self.ffi.new("double *")
        self.handle_return(self.lib.AT_GetFloatMin(AT_H, u(command), result))
        return result[0]

    def get_bool(self, AT_H, command):
        """Run command and get Bool return value.
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_GetBool(AT_H, u(command), result))
        return result[0]

    def set_bool(self, AT_H, command, value):
        """Set command with Bool value parameter.
        """
        self.handle_return(self.lib.AT_SetBool(AT_H, u(command), value))

    def set_enum_index(self, AT_H, command, value):
        """Set command with Enumerated value parameter.
        """
        self.handle_return(self.lib.AT_SetEnumIndex(AT_H, u(command), value))

    def set_enum_string(self, AT_H, command, item):
        """Set command with EnumeratedString value parameter.
        """
        self.handle_return(self.lib.AT_SetEnumString(AT_H, u(command), u(item)))

    def get_enum_index(self, AT_H, command):
        """Run command and set Enumerated return value.
        """
        result = self.ffi.new("int *")
        self.handle_return(self.lib.AT_GetEnumIndex(AT_H, u(command), result))
        return result[0]

    def get_enum_string_options(self, AT_H, command) :
        """Get list of option strings
        """
        count = self.get_enum_count(AT_H, command)
        strings = []
        for i in range(0, count):
            strings.append(self.get_enum_string_by_index(AT_H, u(command),i))
        return strings

    def get_enum_string(self, AT_H, command, result_length=128):
        """Run command and set Enumerated return value.
        """
        ret = self.get_enum_index(AT_H, command)
        return self.get_enum_string_by_index(AT_H, u(command), ret)

    def get_enum_count(self, AT_H, command):
        """Run command and set Enumerated return value.
        """
        result = self.ffi.new("int *")
        self.handle_return(self.lib.AT_GetEnumCount(AT_H, u(command), result))
        return result[0]

    def is_enum_index_available(self, AT_H, command, index):
        """Check if enumerated index is available
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_IsEnumIndexAvailable(AT_H, u(command), index, result))
        return result[0]

    def is_enum_index_implemented(self, AT_H, command, index):
        """Check if enumerated index is implemented
        """
        result = self.ffi.new("AT_BOOL *")
        self.handle_return(self.lib.AT_IsEnumIndexImplemented(AT_H, u(command), index, result))
        return result[0]

    def get_enum_string_by_index(self, AT_H, command, index, result_length=128):
        """Get command with EnumeratedString value parameter.
        """
        result = self.ffi.new("AT_WC [%s]" % result_length)
        self.handle_return(self.lib.AT_GetEnumStringByIndex(AT_H, u(command), index, result, result_length))
        return self.ffi.string(result)

    def command(self, AT_H, command):
        """Run command.
        """
        self.handle_return(self.lib.AT_Command(AT_H, u(command)))

    def set_string(self, AT_H, command, strvalue):
        """SetString function.
        """
        self.handle_return(self.lib.AT_SetString(AT_H, u(command), u(strvalue)))

    def get_string(self, AT_H, command, result_length=128):
        """Run command and get string return value.
        """
        result = self.ffi.new("AT_WC [%s]" % result_length)
        self.handle_return(self.lib.AT_GetString(AT_H, u(command), result, result_length))
        return self.ffi.string(result, result_length)

    def get_string_max_length(self, AT_H, command):
        """Run command and get maximum Int return value.
        """
        result = self.ffi.new("int *")
        self.handle_return(self.lib.AT_GetStringMaxLength(AT_H, u(command), result))
        return result[0]

    def queue_buffer(self, AT_H, buf_ptr, buffer_size):
        """Put buffer in queue.
        """
        self.handle_return(self.lib.AT_QueueBuffer(AT_H, self.ffi.cast("AT_U8 *", buf_ptr), buffer_size))

    def wait_buffer(self, AT_H, timeout=20000):
        """Wait for next buffer to fill.
        """
        buf_ptr = self.ffi.new("AT_U8 **")
        buffer_size = self.ffi.new("int *")
        self.handle_return(self.lib.AT_WaitBuffer(AT_H, buf_ptr, buffer_size, int(timeout)))
        return (buf_ptr, buffer_size[0])

    def flush(self, AT_H):
        self.handle_return(self.lib.AT_Flush(AT_H))

